/*
 *  ofxOmxRpiVideoPlayer.cpp
 *  openFrameworksLib
 *
 *  Created by jason van cleave on 2/7/13.
 *  Copyright 2013 jasonvancleave.com. All rights reserved.
 *
 */

#include "ofxOmxRpiVideoPlayer.h"

#define OMX_INIT_STRUCTURE(a) \
memset(&(a), 0, sizeof(a)); \
(a).nSize = sizeof(a); \
(a).nVersion.s.nVersionMajor = OMX_VERSION_MAJOR; \
(a).nVersion.s.nVersionMinor = OMX_VERSION_MINOR; \
(a).nVersion.s.nRevision = OMX_VERSION_REVISION; \
(a).nVersion.s.nStep = OMX_VERSION_STEP


COMPONENT_T* video_render;
OMX_BUFFERHEADERTYPE* eglBufferStatic;
void onFillBufferDone(void* data, COMPONENT_T* comp)
{    
	if (OMX_FillThisBuffer(ilclient_get_handle(video_render), eglBufferStatic) != OMX_ErrorNone)
	{
		ofLogError() << "OMX_FillThisBuffer failed in callback";
	}
}


ofxOmxRpiVideoPlayer::ofxOmxRpiVideoPlayer()
{
	doLoopVideo = true;
	
	eglImage = NULL;
	eglBuffer = NULL;
	renderer = NULL;
	CLOCK_OUTPUT_PORT = 80;
	DECODER_INPUT_PORT = 130;
	DECODER_OUTPUT_PORT = 131;
	SCHEDULER_INPUT_PORT = 10;
	SCHEDULER_OUTPUT_PORT = 11;
	SCHEDULER_COMMUNICATION_PORT = 12;
	EGL_RENDER_INPUT_PORT = 220;
	EGL_RENDER_OUTPUT_PORT = 221;
	
	filename = "/opt/vc/src/hello_pi/hello_video/test.h264";
	decoder = NULL;
	scheduler = NULL;
	clock = NULL;
	ILCLIENT_T *client;
	
	file = NULL;
	status = 0;
	data = NULL;
	dataLength = 0;
	find_start_codes = 0;
	packet_size = 16<<10; 
}

void ofxOmxRpiVideoPlayer::setup(int width, int height)
{
	bcm_host_init();
	ofDisableArbTex();

	ofAppEGLWindow *appEGLWindow = (ofAppEGLWindow *) ofGetWindowPtr();
	display = appEGLWindow->getEglDisplay();
	context = appEGLWindow->getEglContext();
	
	textureSource.allocate(width, height, GL_RGBA);
	textureSource.getTextureData().bFlipTexture = true;
	textureSource.setTextureWrap(GL_REPEAT, GL_REPEAT);
	texture = textureSource.getTextureData().textureID;
	
	
	glEnable(GL_TEXTURE_2D);
	
	
	// setup first texture
	int dataSize = width * height * 4;
	
	GLubyte* pixelData = new GLubyte [dataSize];
	
	
    memset(pixelData, 0xff, dataSize);  // white texture, opaque
	
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
				 GL_RGBA, GL_UNSIGNED_BYTE, pixelData);
	
	delete[] pixelData;
	
	
	// Create EGL Image
	eglImage = eglCreateImageKHR(
								 display,
								 context,
								 EGL_GL_TEXTURE_2D_KHR,
								 (EGLClientBuffer)texture,
								 0);
    glDisable(GL_TEXTURE_2D);
	if (eglImage == EGL_NO_IMAGE_KHR)
	{
		ofLogError() << "Could not create EGLImage";
		return;
	}
	else
	{
		ofLogVerbose() << "Created EGLImage";
	}
	initComponents();
	startThread(false, true);
}

int ofxOmxRpiVideoPlayer::initComponents()
{
	
	
	if((file = fopen(filename, "rb")) == NULL)
		return -2;
	
	if((client = ilclient_init()) == NULL)
	{
		fclose(file);
		return -3;
	}
	
	if(OMX_Init() != OMX_ErrorNone)
	{
		ilclient_destroy(client);
		fclose(file);
		return -4;
	}
	data = new unsigned char[packet_size+4];
	if(find_start_codes && data == NULL)
	{
		status = -16;
		if(OMX_Deinit() != OMX_ErrorNone)
			status = -17;
		ilclient_destroy(client);
		fclose(file);
		return status;
	}
	
	// callback
	ilclient_set_fill_buffer_done_callback(client, onFillBufferDone, 0);
	
	
	// create decoder
	if(ilclient_create_component(client, &decoder, "video_decode", (ILCLIENT_CREATE_FLAGS_T)(ILCLIENT_DISABLE_ALL_PORTS | ILCLIENT_ENABLE_INPUT_BUFFERS)) != 0)
		status = -14;
	components[0] = decoder;
	
	// create renderer
	if(status == 0 && ilclient_create_component(client, &renderer, "egl_render", (ILCLIENT_CREATE_FLAGS_T)(ILCLIENT_DISABLE_ALL_PORTS | ILCLIENT_ENABLE_OUTPUT_BUFFERS)) != 0)
		status = -14;
	components[1] = renderer;
	video_render = renderer;
	
	// create clock
	if(status == 0 && ilclient_create_component(client, &clock, "clock", (ILCLIENT_CREATE_FLAGS_T)ILCLIENT_DISABLE_ALL_PORTS) != 0)
		status = -14;
	components[2] = clock;
	
	
	OMX_INIT_STRUCTURE(clockState);
	clockState.eState = OMX_TIME_ClockStateWaitingForStartTime;
	clockState.nWaitMask = 1;
	
	OMX_SetParameter(ilclient_get_handle(clock), OMX_IndexConfigTimeClockState, &clockState);
	
	
	ilclient_create_component(client, &scheduler, "video_scheduler", (ILCLIENT_CREATE_FLAGS_T)ILCLIENT_DISABLE_ALL_PORTS);
	components[3] = scheduler;
	
	set_tunnel(tunnel, decoder, DECODER_OUTPUT_PORT, scheduler, SCHEDULER_INPUT_PORT);
	//set_tunnel(tunnel+1, scheduler, SCHEDULER_OUTPUT_PORT, renderer, 90);
	set_tunnel(tunnel+1, scheduler, SCHEDULER_OUTPUT_PORT, renderer, EGL_RENDER_INPUT_PORT);
	set_tunnel(tunnel+2, clock, CLOCK_OUTPUT_PORT, scheduler, SCHEDULER_COMMUNICATION_PORT);
	
	// setup clock tunnel first
	ilclient_setup_tunnel(tunnel+2, 0, 0);
	ilclient_change_component_state(clock, OMX_StateExecuting);
	
	
	ilclient_change_component_state(decoder, OMX_StateIdle);
	
	
	OMX_INIT_STRUCTURE(format);
	format.nPortIndex = DECODER_INPUT_PORT;
	format.eCompressionFormat = OMX_VIDEO_CodingAVC;
	return 0;
}

void ofxOmxRpiVideoPlayer::threadedFunction()
{
	while (isThreadRunning()) 
	{
		if(status == 0 &&
		   OMX_SetParameter(ilclient_get_handle(decoder), OMX_IndexParamVideoPortFormat, &format) == OMX_ErrorNone &&
		   ilclient_enable_port_buffers(decoder, DECODER_INPUT_PORT, NULL, NULL, NULL) == 0)
		{
			OMX_BUFFERHEADERTYPE *inputBuffer;
			int havePortSettingsChanged = 0;
			int isFirstPacket = 1;
			
			ilclient_change_component_state(decoder, OMX_StateExecuting);
			
			while((inputBuffer = ilclient_get_input_buffer(decoder, DECODER_INPUT_PORT, 1)) != NULL)
			{
				// feed data and wait until we get port settings changed
				unsigned char *destination;
				if (find_start_codes)
				{
					destination = data + dataLength;
				}else 
				{
					destination = inputBuffer->pBuffer;
				}
				
				
				// loop if at end
				if (doLoopVideo) 
				{
					if (feof(file))
					{
						rewind(file);
					}
				}
				
				
				dataLength += fread(destination, 1, packet_size+(find_start_codes*4)-dataLength, file);
				
				if(havePortSettingsChanged == 0 &&
				   ((dataLength > 0 && ilclient_remove_event(decoder, OMX_EventPortSettingsChanged, DECODER_OUTPUT_PORT, 0, 0, 1) == 0) ||
					(dataLength == 0 && ilclient_wait_for_event(decoder, OMX_EventPortSettingsChanged, DECODER_OUTPUT_PORT, 0, 0, 1,
																ILCLIENT_EVENT_ERROR | ILCLIENT_PARAMETER_CHANGED, 10000) == 0)))
				{
					havePortSettingsChanged = 1;
					
					if(ilclient_setup_tunnel(tunnel, 0, 0) != 0)
					{
						ofLogError() << "tunnel to scheduler failed";
						break;
					}
					
					ilclient_change_component_state(scheduler, OMX_StateExecuting);
					
					// now setup tunnel to renderer
					if(ilclient_setup_tunnel(tunnel+1, 0, 1000) != 0)
					{
						ofLogError() << "tunnel to renderer failed";
						break;
					}
					
					
					// Set egl_render to idle
					ilclient_change_component_state(renderer, OMX_StateIdle);
					
					
					// Enable the output port and tell egl_render to use the texture as a buffer
					if (OMX_SendCommand(ilclient_get_handle(renderer), OMX_CommandPortEnable, EGL_RENDER_OUTPUT_PORT, NULL) != OMX_ErrorNone)
					{
						ofLogError() << "Enabled egl_render output port FAIL";
						return;
					}else 
					{
						ofLogVerbose() << "Enabled egl_render PASS";
					}

					
					if (OMX_UseEGLImage(ilclient_get_handle(renderer), &eglBuffer, EGL_RENDER_OUTPUT_PORT, NULL, eglImage) != OMX_ErrorNone)
					{
						ofLogError() << "Set renderer to use OMX_UseEGLImage FAIL";
						return;
					}else 
					{
						ofLogVerbose() << "Set renderer to use OMX_UseEGLImage PASS";
						eglBufferStatic = eglBuffer;
					}

					
					
					// Set egl_render to executing
					ilclient_change_component_state(renderer, OMX_StateExecuting);
					
					
					// Request egl_render to write data to the texture buffer
					if(OMX_FillThisBuffer(ilclient_get_handle(renderer), eglBuffer) != OMX_ErrorNone)
					{
						ofLogError() << "OMX_FillThisBuffer FAIL";
						return;
					}else
					{
						ofLogVerbose() << "OMX_FillThisBuffer PASS";
					}

				}
				if(!dataLength)
				{
					break;
				}
				
				if(find_start_codes)
				{
					int i;
					int start = -1;
					int length = 0;
					int maxLength = dataLength > packet_size ? packet_size : dataLength;
					for(i=2; i<maxLength; i++)
					{
						if(data[i-2] == 0 && data[i-1] == 0 && data[i] == 1)
						{
							length = 3;
							start = i-2;
							
							// check for 4 byte start code
							if(i > 2 && data[i-3] == 0)
							{
								length++;
								start--;
							}
							
							break;
						}
					}
					
					if(start == 0)
					{
						// start code is next, so just send that
						inputBuffer->nFilledLen = length;
					} 
					else if(start == -1)
					{
						// no start codes seen, send the first block
						inputBuffer->nFilledLen = maxLength;
					}else
					{
						// start code in the middle of the buffer, send up to the code
						inputBuffer->nFilledLen = start;
					}
					
					memcpy(inputBuffer->pBuffer, data, inputBuffer->nFilledLen);
					memmove(data, data + inputBuffer->nFilledLen, dataLength - inputBuffer->nFilledLen);
					dataLength -= inputBuffer->nFilledLen;
				}else
				{
					inputBuffer->nFilledLen = dataLength;
					dataLength = 0;
				}
				
				inputBuffer->nOffset = 0;
				if(isFirstPacket)
				{
					inputBuffer->nFlags = OMX_BUFFERFLAG_STARTTIME;
					isFirstPacket = 0;
				}else
				{
					inputBuffer->nFlags = OMX_BUFFERFLAG_TIME_UNKNOWN;
				}
				
				if(OMX_EmptyThisBuffer(ilclient_get_handle(decoder), inputBuffer) != OMX_ErrorNone)
				{
					status = -6;
					break;
				}
				
			}
			
			inputBuffer->nFilledLen = 0;
			inputBuffer->nFlags = OMX_BUFFERFLAG_TIME_UNKNOWN | OMX_BUFFERFLAG_EOS;
			
			if(OMX_EmptyThisBuffer(ilclient_get_handle(decoder), inputBuffer) != OMX_ErrorNone)
			{
				ofLogError() << "OMX_EmptyThisBuffer FAIL";
			}
			
			// wait for EOS (End of Stream) from render
			ilclient_wait_for_event(renderer, OMX_EventBufferFlag, 90, 0, OMX_BUFFERFLAG_EOS, 0,
									ILCLIENT_BUFFER_FLAG_EOS, 10000);
			
			// need to flush the renderer to allow decoder to disable its input port
			ilclient_flush_tunnels(tunnel, 0);
			
			ilclient_disable_port_buffers(decoder, DECODER_INPUT_PORT, NULL, NULL, NULL);
		}
		closeComponents();
		
	}
}

void ofxOmxRpiVideoPlayer::closeComponents()
{
	fclose(file);
	
	ilclient_disable_tunnel(tunnel);
	ilclient_disable_tunnel(tunnel+1);
	ilclient_disable_tunnel(tunnel+2);
	ilclient_teardown_tunnels(tunnel);
	
	ilclient_state_transition(components, OMX_StateIdle);
	ilclient_state_transition(components, OMX_StateLoaded);
	
	ilclient_cleanup_components(components);
	
	OMX_Deinit();
	
	ilclient_destroy(client);
	
	
}
ofxOmxRpiVideoPlayer::~ofxOmxRpiVideoPlayer()
{
	
	if (eglImage !=NULL) 
	{
		eglDestroyImageKHR(display, eglImage);
	}
	
}